"""
This file contains functions that might be generally useful.
"""

import contextlib
import datetime
import functools
import os
import logging
import time
import warnings
import webtest  # importing the library makes it easier to mock testing


# Is this the right place for this? I feel like this should be done in an application, not a library.
# -kmp 27-Apr-2020
logging.basicConfig()


# Using PRINT(...) for debugging, rather than its more familiar lowercase form) for intended programmatic output,
# makes it easier to find stray print statements that were left behind in debugging. -kmp 30-Mar-2020

PRINT = print


class VirtualAppError(Exception):
    """ Special Exception to be raised by VirtualApp that contains some additional info """

    def __init__(self, msg, url, body, raw_exception):
        super(VirtualAppError, self).__init__(msg)
        self.msg = msg
        self.query_url = url
        self.query_body = body
        self.raw_exception = raw_exception

    def __repr__(self):
        return ("Exception encountered on VirtualApp\n"
                "URL: %s\n"
                "BODY: %s\n"
                "MSG: %s\n"
                "Raw Exception: %s\n" % (self.query_url, self.query_body, self.msg, self.raw_exception))

    def __str__(self):
        return self.__repr__()


class _VirtualAppHelper(webtest.TestApp):  # effectively disguises 'TestApp'
    pass


class VirtualApp:
    """
    Wrapper class for TestApp, to allow custom control over submitting Encoded requests,
    simulating a number of conditions, including permissions.

    IMPORTANT: We use webtest.TestApp is used as substrate technology here, but use of this class
        occurs in the main application, not just in testing. Among other things, we have
        renamed the app here in order to avoid confusions created by the name when it is used
        in production settings.
    """
    HELPER_CLASS = _VirtualAppHelper

    def __init__(self, app, environ):
        """
        Builds an encoded application, allowing you to submit requests to an encoded application

        :param app: return value of get_app(config_uri, app_name)
        :param environ: options to pass to the application. Usually permissions.
        """
        #  NOTE: The TestApp class that we're wrapping takes a richer set of initialization parameters
        #        (including relative_to, use_unicode, cookiejar, parser_features, json_encoder, and lint),
        #        but we'll add them conservatively here. If there is a need for any of them, we should add
        #        them explicitly here one-by-one as the need is shown so we have tight control of what
        #        we're depending on and what we're not. -kmp 27-Apr-2020
        self.wrapped_app = self.HELPER_CLASS(app, environ)

    def get(self, url, **kwargs):
        """ Wrapper for TestApp.get that logs the outgoing GET

        :param url: url to GET
        :param kwargs: args to pass to the GET
        :return: result of GET
        """
        logging.info('OUTGOING HTTP GET: %s' % url)
        try:
            return self.wrapped_app.get(url, **kwargs)
        except webtest.AppError as e:
            raise VirtualAppError(msg='HTTP GET failed.', url=url, body='<empty>', raw_exception=e)

    def post_json(self, url, obj, **kwargs):
        """ Wrapper for TestApp.post_json that logs the outgoing POST

        :param url: url to POST to
        :param obj: object body to POST
        :param kwargs: args to pass to the POST
        :return: result of POST
        """
        logging.info('OUTGOING HTTP POST on url: %s with object: %s' % (url, obj))
        try:
            return self.wrapped_app.post_json(url, obj, **kwargs)
        except webtest.AppError as e:
            raise VirtualAppError(msg='HTTP POST failed.', url=url, body=obj, raw_exception=e)

    def put_json(self, url, obj, **kwargs):
        """ Wrapper for TestApp.put_json that logs the outgoing PUT

        :param url: url to PUT to
        :param obj: object body to PUT
        :param kwargs: args to pass to the PUT
        :return: result of PUT
        """
        logging.info('OUTGOING HTTP PUT on url: %s with object: %s' % (url, obj))
        try:
            return self.wrapped_app.put_json(url, obj, **kwargs)
        except webtest.AppError as e:
            raise VirtualAppError(msg='HTTP PUT failed.', url=url, body=obj, raw_exception=e)

    def patch_json(self, url, fields, **kwargs):
        """ Wrapper for TestApp.patch_json that logs the outgoing PATCH

        :param url: url to PATCH to, should contain an object uuid
        :param fields: fields to PATCH on uuid in URL
        :param kwargs: args to pass to the PATCH
        :return: result of PATCH
        """
        logging.info('OUTGOING HTTP PATCH on url: %s with changes: %s' % (url, fields))
        try:
            return self.wrapped_app.patch_json(url, fields, **kwargs)
        except webtest.AppError as e:
            raise VirtualAppError(msg='HTTP PATCH failed.', url=url, body=fields, raw_exception=e)

    @property
    def app(self):
        """ Returns the .app of the wrapped_app.

            For example, this allows one to refer to myapp.app.registry without having to know
            if myapp is a TestApp or a VirtualApp.
        """
        return self.wrapped_app.app


def ignored(*args, **kwargs):
    """
    This is useful for defeating flake warnings.
    Call this function to use values that really should be ignored.
    This is intended as a declaration that variables are intentionally ignored,
    but no enforcement of that is done. Some sample uses:

    def foo(x, y):
        ignored(x, y)  # so flake8 won't complain about x and y being unused.
        return 3

    def map_action(action, data, options, precheck=ignored):
        precheck(data, **options)
        action(data, **options)
    """
    return args, kwargs


def get_setting_from_context(settings, ini_var, env_var=None, default=None):
    """
    This gets a value from either an environment variable or a config file.

    The environment variable overrides, since it is more dynamic in nature than a config file,
    which might be checked into source control.

    If the value of env_var is None, it will default to a name similar to ini_var,
    but in uppercase and with '.' replaced by '_'. So a 'foo.bar' ini file setting
    will defaultly correspond to a 'FOO_BAR' environment variable. This can be overridden
    by using an string argument for env_var to specify the environment variable, or using False
    to indicate that no env_var is allowed.
    """
    if env_var is not False:  # False specially means don't allow an environ variable, in case that's ever needed.
        if env_var is None:
            # foo.bar.baz in config file corresponds to FOO_BAR_BAZ as an environment variable setting.
            env_var = ini_var.upper().replace(".", "_")
        # NOTE WELL: An implication of this is that an environment variable of an empty string
        #            will override a config file setting that is non-empty. This uses 'principle of least surprise',
        #            that if environment variable settings appear to set a null string, that's what should prevail.
        if env_var in os.environ:
            return os.environ.get(env_var)
    return settings.get(ini_var, default)


@contextlib.contextmanager
def filtered_warnings(action, message="", category=None, module="", lineno=0, append=False):
    """
    Context manager temporarily filters deprecation messages for the duration of the body.

    Except for its dynamic scope, this is used otherwise the same as warnings.filterwarnings would be used.

    If category is unsupplied, it should be a class object that is Warning (the default) or one of its subclasses.

    For example:

           with filtered_warnings('ignore', category=DeprecationWarning):
               ... use something that's deprecated without a lot of fuss ...

    Note: This is not threadsafe. It's OK while loading system and during testing,
          but not in worker threads.
    """
    if category is None:
        category = Warning
    with warnings.catch_warnings():
        warnings.filterwarnings(action, message=message, category=category, module=module,
                                lineno=lineno, append=append)
        yield


class Retry:

    """
    This class exists primarily to hold onto data relevant to the Retry.retry_allowed decorator.
    There is no need to instantiate the class in order for it to work.

    This class also has a subclass qa_utils.RetryManager that adds the ability to locally bind data
    that has been declared with this decorator.
    """

    class RetryOptions:
        """
        A helper class used internally by the Retry class.

        One of these objects is created and registered for each decorated function unless the name_key is 'anonymous'.
        See Retry._RETRY_OPTIONS_CATALOG.
        """

        def __init__(self, retries_allowed=None, wait_seconds=None, wait_increment=None, wait_multiplier=None):
            self.retries_allowed = retries_allowed
            self.wait_seconds = wait_seconds or 0  # None or False mean 0 seconds
            self.wait_increment = wait_increment
            self.wait_multiplier = wait_multiplier
            self.wait_adjustor = self.make_wait_adjustor(wait_increment=wait_increment, wait_multiplier=wait_multiplier)

        @staticmethod
        def make_wait_adjustor(wait_increment=None, wait_multiplier=None):
            """
            Returns a function that can be called to adjust wait_seconds based on wait_increment or wait_multiplier
            before doing a retry at each step.
            """
            if wait_increment and wait_multiplier:
                raise SyntaxError("You may not specify both wait_increment and wait_multiplier.")

            if wait_increment:
                return lambda x: x + wait_increment
            elif wait_multiplier:
                return lambda x: x * wait_multiplier
            else:
                return lambda x: x

        @property
        def tries_allowed(self):
            return 1 + self.retries_allowed

    _RETRY_OPTIONS_CATALOG = {}

    DEFAULT_RETRIES_ALLOWED = 1
    DEFAULT_WAIT_SECONDS = 0
    DEFAULT_WAIT_INCREMENT = None
    DEFAULT_WAIT_MULTIPLIER = None

    @classmethod
    def _defaulted(cls, value, default):
        """ Triages between argument values and class-declared defaults. """
        return default if value is None else value

    @classmethod
    def retry_allowed(cls, name_key=None, retries_allowed=None, wait_seconds=None,
                      wait_increment=None, wait_multiplier=None):
        """
        Used as a decorator on a function definition, makes that function do retrying before really failing.
        For example:

            @Retry.retry_allowed(retries_allowed=4, wait_seconds=2, wait_multiplier=1.25)
            def something_that_fails_a_lot(...):
                ... flaky code ...

        will cause the something_that_fails_a_lot(...) code to retry several times before giving up,
        either using the same wait each time or, if given a wait_multiplier or wait_increment, using
        that advice to adjust the wait time upward on each time.

        Args:

            name_key: An optional key that can be used by qa_utils.RetryManager to adjust these parameters in testing.
                      If the argument is 'anonymous', no record will be created.
            retries_allowed: The number of retries allowed. Default is cls.DEFAULT_RETRIES_ALLOWED.
            wait_seconds: The number of wait_seconds between retries. Default is cls.DEFAULT_WAIT_SECONDS.
            wait_increment: A fixed increment by which the number of wait_seconds is adjusted on each retry.
            wait_multiplier: A multiplier by which the number of wait_seconds is adjusted on each retry.
        """

        def decorator(function):
            function_name = name_key or function.__name__
            function_profile = cls.RetryOptions(
                retries_allowed=cls._defaulted(retries_allowed, cls.DEFAULT_RETRIES_ALLOWED),
                wait_seconds=cls._defaulted(wait_seconds, cls.DEFAULT_WAIT_SECONDS),
                wait_increment=cls._defaulted(wait_increment, cls.DEFAULT_WAIT_INCREMENT),
                wait_multiplier=cls._defaulted(wait_multiplier, cls.DEFAULT_WAIT_MULTIPLIER),
            )

            check_true(isinstance(retries_allowed, int) and retries_allowed >= 0,
                       "The retries_allowed must be a non-negative integer.",
                       error_class=ValueError)

            # See the 'retrying' method to understand what this is about. -kmp 8-Jul-2020
            if function_name != 'anonymous':
                cls._RETRY_OPTIONS_CATALOG[function_name] = function_profile  # Only for debugging.

            @functools.wraps(function)
            def wrapped_function(*args, **kwargs):
                tries_allowed = function_profile.tries_allowed
                wait_seconds = function_profile.wait_seconds or 0
                last_error = None
                for i in range(tries_allowed):
                    if i > 0:
                        if i > 1:
                            wait_seconds = function_profile.wait_adjustor(wait_seconds)
                        if wait_seconds > 0:
                            time.sleep(wait_seconds)
                    try:
                        success = function(*args, **kwargs)
                        return success
                    except Exception as e:
                        last_error = e
                if last_error is not None:
                    raise last_error

            return wrapped_function

        return decorator

    @classmethod
    def retrying(cls, fn, retries_allowed=None, wait_seconds=None, wait_increment=None, wait_multiplier=None):
        """
        Similar to the @Retry.retry_allowed decorator, but used around individual calls. e.g.,

            res = Retry.retrying(testapp.get)(url)

        If you don't like the defaults, you can override them with arguments:

            res = Retry.retrying(testapp.get, retries_allowed=5, wait_seconds=1)(url)

        but if you need to do it a lot, you can make a subclass:

            class MyRetry(Retry):
                DEFAULT_RETRIES_ALLOWED = 5
                DEFAULT_WAIT_SECONDS = 1
            retrying = MyRetry.retrying  # Avoids saying MyRetry.retrying(...) everywhere
            ...
            res1 = retrying(testapp.get)(url)
            res2 = retrying(testapp.get)(url)
            ...etc.

        Args:

            fn: A function that will be retried on failure.
            retries_allowed: The number of retries allowed. Default is cls.DEFAULT_RETRIES_ALLOWED.
            wait_seconds: The number of wait_seconds between retries. Default is cls.DEFAULT_WAIT_SECONDS.
            wait_increment: A fixed increment by which the number of wait_seconds is adjusted on each retry.
            wait_multiplier: A multiplier by which the number of wait_seconds is adjusted on each retry.

        Returns: whatever the fn returns, assuming it returns normally/successfully.
        """
        # A special name_key of 'anonymous' is the default, which causes there not to be a name key.
        # This cannot work in conjunction with RetryManager because different calls may result in different
        # function values at the same point in code. -kmp 8-Jul-2020
        decorator_function = Retry.retry_allowed(
            name_key='anonymous', retries_allowed=retries_allowed, wait_seconds=wait_seconds,
            wait_increment=wait_increment, wait_multiplier=wait_multiplier
        )
        return decorator_function(fn)


def apply_dict_overrides(dictionary: dict, **overrides) -> dict:
    """
    Assigns a given set of overrides to a dictionary, ignoring any entries with None values, which it leaves alone.
    """
    # I'm not entirely sure the treatment of None is the right thing. Need to look into that.
    # Then again, if None were stored, then apply_dict_overrides(d, var1=1, var2=2, var3=None)
    # would be no different than (dict(d, var1=1, var2=2, var3=None). It might be more useful
    # and/or interesting if it would actually remove the key instead. -kmp 18-Jul-2020
    for k, v in overrides.items():
        if v is not None:
            dictionary[k] = v
    # This function works by side effect, but getting back the changed dict may be sometimes useful.
    return dictionary


def utc_today_str():
    """Returns a YYYY-mm-dd date string, relative to the UTC timezone."""
    return datetime.datetime.strftime(datetime.datetime.utcnow(), "%Y-%m-%d")


class LockoutManager:
    """
    This class is used as a guard of a critical operation that can only be called within a certain frequency.
    e.g.,

        class Foo:
            def __init__(self):
                # 60 seconds required between calls, with a 1 second margin of error (overhead, clocks varying, etc)
                self.lockout_manager = LockoutManager(action="foo", lockout_seconds=1, safety_seconds=1)
            def foo():
                self.lockout_manager.wait_if_needed()
                do_guarded_action()

        f = Foo()     # make a Foo
        v1 = f.foo()  # will immediately get a value
        time.sleep(58)
        v2 = f.foo()  # will wait about 2 seconds, then get a value
        v3 = f.foo()  # will wait about 60 seconds, then get a value

    Conceptually this is a special case of RateManager for n=1, though in practice it arose differently and
    the supplementary methods (which we happen to use mostly for testing) differ because the n=1 case is simpler
    and admits more questions. So, for now at least, this is not a subclass of RateManager but a separate
    implementation.
    """

    EARLIEST_TIMESTAMP = datetime.datetime(datetime.MINYEAR, 1, 1)  # maybe useful for testing

    def __init__(self, *, lockout_seconds, safety_seconds=0, action="metered action", enabled=True, log=None):
        """
        Creates a LockoutManager that cooperates in assuring a guarded operation is only happens at a certain rate.

        The rate is once person lockout_seconds. This is a special case of RateManager and might get phased out
        as redundant, but has slightly different operations available for testing.

        Args:

        lockout_seconds int: A theoretical number of seconds allowed between calls to the guarded operation.
        safety_seconds int: An amount added to interval_seconds to accommodate real world coordination fuzziness.
        action str: A noun or noun phrase describing the action being guarded.
        enabled bool: A boolean controlling whether this facility is enabled. If False, waiting is disabled.
        log object: A logger object (supporting operations like .debug, .info, .warning, and .error).
        """

        # This makes it easy to turn off the feature
        self.lockout_enabled = enabled
        self.lockout_seconds = lockout_seconds
        self.safety_seconds = safety_seconds
        self.action = action
        self._timestamp = self.EARLIEST_TIMESTAMP
        self.log = log or logging

    @property
    def timestamp(self):
        """The timestamp is read-only. Use update_timestamp() to set it."""
        return self._timestamp

    @property
    def effective_lockout_seconds(self):
        """
        The effective time between calls

        Returns: the sum of the lockout and the safety seconds
        """
        return self.lockout_seconds + self.safety_seconds

    def wait_if_needed(self):
        """
        This function is intended to be called immediately prior to each guarded operation.

        This function will wait (using time.sleep) only if necessary, and for the amount necessary,
        to comply with rate-limiting declared in the creation of this LockoutManager.

        NOTE WELL: It is presumed that all calls are coming from this source. This doesn't have ESP that would
        detect or otherwise accommodate externally generated calls, so violations of rate-limiting can still
        happen that way. This should be sufficient for sequential testing, and better than nothing for
        production operation.  This is not a substitute for responding to server-initiated throttling protocols.
        """
        now = datetime.datetime.now()
        # Note that this quantity is always positive because now is always bigger than the timestamp.
        seconds_since_last_attempt = (now - self._timestamp).total_seconds()
        # Note again that because seconds_since_last_attempt is positive, the wait seconds will
        # never exceed self.effective_lockout_seconds, so
        #   0 <= wait_seconds <= self.effective_lockout_seconds
        wait_seconds = max(0.0, self.effective_lockout_seconds - seconds_since_last_attempt)
        if wait_seconds > 0.0:
            shared_message = ("Last %s attempt was at %s (%s seconds ago)."
                              % (self.action, self._timestamp, seconds_since_last_attempt))
            if self.lockout_enabled:
                action_message = "Waiting %s seconds before attempting another." % wait_seconds
                self.log.warning("%s %s" % (shared_message, action_message))
                time.sleep(wait_seconds)
            else:
                action_message = "Continuing anyway because lockout is disabled."
                self.log.warning("%s %s" % (shared_message, action_message))
        self.update_timestamp()

    def update_timestamp(self):
        """
        Explicitly sets the reference time point for computation of our lockout.
        This is called implicitly by .wait_if_needed(), and for some situations that may be sufficient.
        """
        self._timestamp = datetime.datetime.now()


class RateManager:
    """
    This class is used for functions that can only be called at a certain rate, described by calls per unit time.
    e.g.,

        class Foo:
            def __init__(self):
                # 60 seconds required between calls, with a 1 second margin of error (overhead, clocks varying, etc)
                self.rate_manager = RateManager(action="foo", interval_seconds=1, safety_seconds=1)
            def foo():
                self.lockout_manager.wait_if_needed()
                do_guarded_action()

        f = Foo()     # make a Foo
        v1 = f.foo()  # will immediately get a value
        time.sleep(58)
        v2 = f.foo()  # will wait about 2 seconds, then get a value
        v3 = f.foo()  # will wait about 60 seconds, then get a value

    Conceptually this is a special case of RateManager for n=1, though in practice it arose differently and
    the supplementary methods (which we happen to use mostly for testing) differ because the n=1 case is simpler
    and admits more questions. So, for now at least, this is not a subclass of RateManager but a separate
    implementation.
    """

    EARLIEST_TIMESTAMP = datetime.datetime(datetime.MINYEAR, 1, 1)  # maybe useful for testing

    def __init__(self, *, interval_seconds, safety_seconds=0, allowed_attempts=1,
                 action="metered action", enabled=True, log=None, wait_hook=None):
        """
        Creates a RateManager that cooperates in assuring that a guarded operation happens only at a certain rate.

        The rate is measured as allowed_attempts per interval_seconds.

        Args:

        interval_seconds int: A number of seconds (the theoretical denominator of the allowed rate)
        safety_seconds int: An amount added to interval_seconds to accommodate real world coordination fuzziness.
        allowed_attempts int: A number of attempts allowed for every interval_seconds.
        action str: A noun or noun phrase describing the action being guarded.
        enabled bool: A boolean controlling whether this facility is enabled. If False, waiting is disabled.
        log object: A logger object (supporting operations like .debug, .info, .warning, and .error).
        wait_hook: A hook not recommended for production, but intended for testing to know when waiting happens.

        """
        if not (isinstance(allowed_attempts, int) and allowed_attempts >= 1):
            raise TypeError("The allowed_attempts must be a positive integer: %s" % allowed_attempts)
        # This makes it easy to turn off the feature
        self.enabled = enabled
        self.interval_seconds = interval_seconds
        self.safety_seconds = safety_seconds
        self.allowed_attempts = allowed_attempts
        self.action = action
        self.timestamps = [self.EARLIEST_TIMESTAMP] * allowed_attempts
        self.log = log or logging
        self.wait_hook = wait_hook

    def set_wait_hook(self, wait_hook):
        """
        Use this to set the wait hook, which will be a function that notices we had to wait.

        Args:

        wait_hook: a function of two arguments (wait_seconds and next_expiration)
        """
        self.wait_hook = wait_hook

    def wait_if_needed(self):
        """
        This function is intended to be called immediately prior to each guarded operation.

        This function will wait (using time.sleep) only if necessary, and for the amount necessary,
        to comply with rate-limiting declared in the creation of this RateManager.

        NOTE WELL: It is presumed that all calls are coming from this source. This doesn't have ESP that would
        detect or otherwise accommodate externally generated calls, so violations of rate-limiting can still
        happen that way. This should be sufficient for sequential testing, and better than nothing for
        production operation.  This is not a substitute for responding to server-initiated throttling protocols.
        """
        now = datetime.datetime.now()
        expiration_delta = datetime.timedelta(seconds=self.interval_seconds)
        latest_expiration = now + expiration_delta
        soonest_expiration = latest_expiration
        # This initial value of soonest_expiration_pos is arbitrarily chosen, but it will normally be superseded.
        # The only case where it's not overridden is where there were no better values than the latest_expiration,
        # so if we wait that amount, all of the slots will be ready to be reused and we might as well use 0 as any.
        # -kmp 19-Jul-2020
        soonest_expiration_pos = 0
        for i, expiration_time in enumerate(self.timestamps):
            if expiration_time <= now:  # This slot was unused or has expired
                self.timestamps[i] = latest_expiration
                return
            elif expiration_time <= soonest_expiration:
                soonest_expiration = expiration_time
                soonest_expiration_pos = i
        sleep_time_needed = (soonest_expiration - now).total_seconds() + self.safety_seconds
        if self.enabled:
            if self.wait_hook:  # Hook primarily for testing
                self.wait_hook(wait_seconds=sleep_time_needed, next_expiration=soonest_expiration)
            self.log.warning("Waiting %s seconds before attempting %s." % (sleep_time_needed, self.action))
            time.sleep(sleep_time_needed)
        # It will have expired now, so grab that slot. We have to recompute the 'now' time because we slept in between.
        self.timestamps[soonest_expiration_pos] = datetime.datetime.now() + expiration_delta


def environ_bool(var, default=False):
    """
    Returns True if the named environment variable is set to 'true' (in any alphabetic case), False if something else.

    If the variable value is not set, the default is returned. False is the default default.
    This function is intended to allow boolean parameters to be initialized from environment variables.
    e.g.,
        DEBUG_FOO = environ_bool("FOO")
    or. if a special value is desired when the variable is not set:
        DEBUG_FOO = environ_bool("FOO", default=None)

    Args:
        var str: The name of an environment variable.
        default object: Any object.
    """
    if var not in os.environ:
        return default
    else:
        return os.environ[var].lower() == "true"


def check_true(test_value, message, error_class=None):
    """
    If the first argument does not evaluate to a true value, an error is raised.

    The error, if one is raised, will be of type error_class, and its message will be given by message.
    The error_class defaults to RuntimeError, but may be any Exception class.
    """
    if error_class is None:
        error_class = RuntimeError
    if not test_value:
        raise error_class(message)


def remove_prefix(prefix, text, required=False):
    if not text.startswith(prefix):
        if required:
            raise ValueError('Prefix %s is not the initial substring of %s' % (prefix, text))
        else:
            return text
    return text[len(prefix):]


def remove_suffix(suffix, text, required=False):
    if not text.endswith(suffix):
        if required:
            raise ValueError('Suffix %s is not the final substring of %s' % (suffix, text))
        else:
            return text
    return text[:len(text)-len(suffix)]



from .misc_utils import PRINT


def _ask_boolean_question(question, quick=None, default=None):
    """
    Loops asking a question interactively until it gets a 'yes' or 'no' response. Returns True or False accordingly.

    :param question: The question to ask (without prompts for possible responses, which will be added automatically).
    :param quick: Whether to allow short-form responses.
    :param default: Whether to provide a default obtained by just pressing Enter.
    :return: True or False
    """

    if quick is None:
        # If the default is not None, we're accepting Enter for yes, so we might as well accept 'y'.
        quick = (default is not None)

    affirmatives = ['y', 'yes'] if quick else ['yes']
    negatives = ['n', 'no'] if quick else ['no']
    affirmative = affirmatives[0]
    negative = negatives[0]
    prompt = ("%s [%s/%s]: "
              % (question,
                 affirmative.upper() if default is True else affirmative,
                 negative.upper() if default is False else negative))
    while True:
        answer = input(prompt).strip().lower()
        if answer in affirmatives:
            return True
        elif answer in negatives:
            return False
        elif answer == "" and default is not None:
            return default
        else:
            PRINT("Please answer '%s' or '%s'." % (affirmative, negative))
            if default is not None:
                PRINT("The default if you just press Enter is '%s'."
                      % (affirmative if default else negative))


def yes_or_no(question):
    """
    Asks a 'yes' or 'no' question.

    Either 'y' or 'yes' (in any case) is an acceptable affirmative.
    Either 'n' or 'no' (in any case) is an acceptable negative.
    The response must be confirmed by pressing Enter.
    There is no default. If Enter is pressed after other text than the valid responses, the user is reprompted.
    """
    return _ask_boolean_question(question, quick=False)


def y_or_n(question, default=None):
    """
    Asks a 'y' or 'n' question.

    Either 'y' or 'yes' (in any case) is an acceptable affirmative.
    Either 'n' or 'no' (in any case) is an acceptable negative.
    The response must be confirmed by pressing Enter.
    If Enter is pressed with no input text, the default is returned if there is one, or else the user is re-prompted.
    If Enter is pressed after other text than the valid responses, the user is reprompted.
    """
    return _ask_boolean_question(question, quick=True, default=default)



import logging
import structlog
import datetime
import uuid
from threading import Timer
from dcicutils import es_utils
from structlog.threadlocal import wrap_dict
from elasticsearch import helpers


class ElasticsearchHandler(logging.Handler):
    """
    Custom handler to post logs to Elasticsearch.
    Needed to sign ES requests with the AWS V4 Signature
    Loosely based off of code here:
    https://github.com/cmanaha/python-elasticsearch-logger
    """
    def __init__(self, env, es_server):
        """
        Must be given a string es_server url to work.
        Calls __init__ of parent Handler
        """
        self.resend_timer = None
        self.messages_to_resend = []
        self.retry_limit = 2
        self.namespace = self.get_namespace(env)
        self.es_client = es_utils.create_es_client(es_server, use_aws_auth=True)
        logging.Handler.__init__(self)

    @staticmethod
    def get_namespace(env):
        """ Grabs ES namespace from health page """
        from .ff_utils import get_health_page
        health = get_health_page(ff_env=env)
        return health.get('namespace', '')

    def schedule_resend(self):
        """
        Create a threading Timer as self.resend_timer to schedule resending any
        records in self.resend_messages after 5 seconds.
        If already resending, do nothing
        """
        if self.resend_timer is None:
            self.resend_timer = Timer(5, self.resend_messages)
            self.resend_timer.daemon = True
            self.resend_timer.start()

    def resend_messages(self):
        """
        Send all records held in self.messages_to_resend in Elasticsearch in bulk.
        Keep track of subsequent errors and retry them, if they have been
        retried fewer times than self.retry_limit
        """
        # clean up the timer
        if self.resend_timer is not None and self.resend_timer.is_alive():
            self.resend_timer.cancel()
        self.resend_timer = None
        if self.messages_to_resend:
            messages_copy = self.messages_to_resend[:]
            self.messages_to_resend = []
            idx_name = self.calculate_log_index()
            actions = (
                {
                    '_index': idx_name,
                    '_type': 'log',
                    '_id': message[0],
                    '_source': message[1]
                }
                for message in messages_copy
            )
            errors = []
            for ok, resp in helpers.streaming_bulk(self.es_client, actions):
                if not ok:
                    errors.append(resp['index']['_id'])
            for sent_message in messages_copy:
                if sent_message[0] in errors and sent_message[2] < self.retry_limit:
                    sent_message[2] += 1  # increment retries
                    self.messages_to_resend.append(sent_message)
        # trigger resending logs if any failed
        if self.messages_to_resend:
            self.schedule_resend()

    def emit(self, record):
        """
        Overload the emit method to post logs to ES
        """
        # required?
        # entry = self.format(record)
        idx_name = self.calculate_log_index()
        # get the message from the record
        message = record.__dict__.get('msg')
        # adds the ability to manually skip logging the message to ES
        if not message or message.get('_skip_es', False) is True:
            return
        # make a uuid; use the inherent log_uuid if provided
        log_id = message.get('log_uuid', str(uuid.uuid4()))
        # for testing purposes. trigger the retry mechanism
        if message.get('_test_log_utils', False) is True:
            self.messages_to_resend.append([log_id, message, 0])
            self.schedule_resend()
            return
        try:
            self.es_client.index(index=idx_name, doc_type='log', body=message, id=log_id)
        except Exception:
            # append resend messages as a list: [<uuid>, <dict message>, <int retries>]
            self.messages_to_resend.append([log_id, message, 0])
            self.schedule_resend()

    def calculate_log_index(self):
        """
        Simple function to name the ES log index by month
        Convention is: logs-<yyyy>-<mm>
        * Uses UTC *
        """
        now = datetime.datetime.utcnow()
        idx_suffix = datetime.datetime.strftime(now, '%Y-%m')
        return self.namespace + 'logs-' + idx_suffix


class ElasticsearchLoggerFactory(structlog.stdlib.LoggerFactory):
    """
    Needed to bind the ElasticsearchHandler to the structlog logger.
    Use for logger_factory arg in structlog.configure function
    See: https://github.com/hynek/structlog/blob/master/src/structlog/stdlib.py
    """
    def __init__(self, env=None, ignore_frame_names=None, es_server=None, in_prod=False):
        """
        Set self.es_server and call __init__ of parent.
        If not in prod, always set the es_server to None (dev mode)
        """
        self.env = None if not env else env
        self.es_server = es_server if in_prod else None
        structlog.stdlib.LoggerFactory.__init__(self, ignore_frame_names)

    def __call__(self, *args):
        """
        Overload the original __call__ function and add the custom handler
        The args used to structlog.getLogger should be <logger name> and
        <es server>, in that order
        """
        if args:
            name = args[0]
        else:
            _, name = structlog._frames._find_first_app_frame_and_name(self._ignore)
        logger = logging.getLogger(name)
        if self.es_server:
            es_handler = ElasticsearchHandler(self.env, self.es_server)
            logger.addHandler(es_handler)
            # also set level to info
            logger.setLevel(logging.INFO)
        return logger


def convert_ts_to_at_ts(logger, log_method, event_dict):
    '''
    this function is used to ensure filebeats
    uses our own timestamp if we logged one
    '''
    if 'timestamp' in event_dict:
        event_dict['@timestamp'] = event_dict['timestamp']
        del event_dict['timestamp']
        return event_dict


def add_log_uuid(logger, log_method, event_dict):
    '''
    this function adds a uuid to the log
    '''
    event_dict['log_uuid'] = str(uuid.uuid4())
    return event_dict


# configure structlog to use its formats for stdlib logging and / or structlog logging
def set_logging(env=None, es_server=None, in_prod=False, level=logging.WARN, log_name=None, log_dir=None):
    '''
    Set logging is a function to be used everywhere, to encourage all subsytems
    to generate structured JSON logs, for easy insertion into ES for searching and
    visualizing later.

    Providing an Elasticsearch server name (es_server) will cause the logs to
    be automatically written to that server. Setting 'skip_es' to True for any
    individual logging statement will cause it not to be written.

    Currently this only JSONifies our own logs, and the bit at the very bottom
    would JSONify other logs, like botocore and stuff but that's  probably more
    than we want to store.

    Also sets some standard handlers for the following:
    add_logger_name - python module generating the log
    timestamper - timestamps... big surprise there
    convert_ts_ta_at_ts - takes our timestamp and overwrides
    the @timestamp key used by filebeats, so queries in ES will
    be against our times, which in things like indexing can differ
    a fair amount from the timestamp inserted by filebeats.
    StackInfoRenderer - capture stack trace and insert into JSON
    format_exc_info - capture exception and insert into JSON
    '''
    timestamper = structlog.processors.TimeStamper(fmt="iso")

    logging.basicConfig(format='')

    # structlog is basically wrapping pythons stdlib logging framework and
    # implementing a custom list of processors that take the log message
    # and do fun stuff with it.  Processors are executed in order of this list
    processors = [
        structlog.stdlib.filter_by_level,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        structlog.stdlib.PositionalArgumentsFormatter(),
        timestamper,
        convert_ts_to_at_ts,
        add_log_uuid,
        structlog.processors.StackInfoRenderer(),
        structlog.processors.format_exc_info,
        structlog.processors.UnicodeDecoder(),
    ]

    # if we are in_prod, do not do any prettifying of the log messages and
    # instead send them to ES. Also set the logging level lower, to INFO
    # in_prod affects the logging handlers used in ElasticsearchLoggerFactory
    if not in_prod:
        # pretty color logs
        processors.append(structlog.dev.ConsoleRenderer())

    # need this guy to go last
    processors.append(structlog.stdlib.ProcessorFormatter.wrap_for_formatter)

    # context_class guy sets context to thread local so we can set logging context from
    # stats-tween and have that included in each subsequent log entry.
    structlog.configure(
        processors=processors,
        context_class=wrap_dict(dict),
        logger_factory=ElasticsearchLoggerFactory(env=env, es_server=es_server, in_prod=in_prod),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )

    # define format and processors for stdlib logging, in case someone hasn't switched
    # yet to using structlog.  Switched off for now as botocore and es are both
    # very chatty, and we should probably turn them to like WARNING level before doing the below
    '''
    pre_chain = [
        # Add the log level and a timestamp to the event_dict if the log entry
        # is not from structlog.
        structlog.stdlib.add_log_level,
        structlog.stdlib.add_logger_name,
        timestamper,
    ]
    format_processor = structlog.dev.ConsoleRenderer()
    if in_prod:
        format_processor = structlog.processors.JSONRenderer()

    formatter = structlog.stdlib.ProcessorFormatter(
        processor=format_processor,
        foreign_pre_chain=pre_chain,
    )

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    root_logger = logging.getLogger()
    root_logger.addHandler(handler)

    root_logger.setLevel(level)
    '''

    # set logging level
    # below could be used ot redirect logging to a file if desired
    if log_name is None:
        log_name = __name__

    if log_dir and log_name and level:
        import os
        log_file = os.path.join(log_dir, log_name + ".log")
        logger = structlog.get_logger(log_name)
        hdlr = logging.FileHandler(log_file)
        formatter = logging.Formatter('')
        hdlr.setFormatter(formatter)
        logger.addHandler(hdlr)
        logger.setLevel(level)
    elif level:
        logger = structlog.get_logger(log_name)
        logger.setLevel(level)



"""
qa_utils: Tools for use in quality assurance testing.
"""

import contextlib
import datetime
import time
import io
import os
import pytz

from .misc_utils import PRINT, ignored, Retry


def show_elapsed_time(start, end):
    """ Helper method for below that is the default - just prints the elapsed time. """
    PRINT('Elapsed: %s' % (end - start))


@contextlib.contextmanager
def timed(reporter=None, debug=None):
    """ A simple context manager that will time how long it spends in context. Useful for debugging.

        :param reporter: lambda x, y where x and y are the start and finish times respectively, default PRINT
        :param debug: lambda x where x is an exception, default NO ACTION
    """
    if reporter is None:
        reporter = show_elapsed_time
    start = time.time()
    try:
        yield
    except Exception as e:
        if debug is not None:
            debug(e)
        raise
    finally:
        end = time.time()
        reporter(start, end)


def mock_not_called(name):
    """
    This can be used in mocking to mock a function that should not be called.
    Called with the name of a function, it returns a function that if called
    will raise an AssertionError complaining that such a name was called.
    """
    def mocked_function(*args, **kwargs):
        # It's OK to print here because we're expected to be called in a testing context, and
        # we're just about to fail a test. The person invoking the tests may want this data.
        PRINT("args=", args)
        PRINT("kwargs=", kwargs)
        raise AssertionError("%s was called where not expected." % name)
    return mocked_function


@contextlib.contextmanager
def local_attrs(obj, **kwargs):
    """
    This binds the named attributes of the given object.
    This is only allowed for an object that directly owns the indicated attributes.

    """
    keys = kwargs.keys()
    for key in keys:
        if key not in obj.__dict__:
            # This works only for objects that directly have the indicated property.
            # That happens for
            #  (a) an instance where its instance variables are in keys.
            #  (b) an uninstantiated class where its class variables (but not inherited class variables) are in keys.
            # So the error happens for these cases:
            #  (c) an instance where any of the keys come from its class instead of the instance itself
            #  (d) an uninstantiated class being used for keys that are inherited rather than direct class variables
            raise ValueError("%s inherits property %s. Treating it as dynamic could affect other objects."
                             % (obj, key))
    saved = {
        key: getattr(obj, key)
        for key in keys
    }
    try:
        for key in keys:
            setattr(obj, key, kwargs[key])
        yield
    finally:
        for key in keys:
            setattr(obj, key, saved[key])


@contextlib.contextmanager
def override_environ(**overrides):
    to_delete = []
    to_restore = {}
    env = os.environ
    try:
        for k, v in overrides.items():
            if k in env:
                to_restore[k] = env[k]
            else:
                to_delete.append(k)
            if v is None:
                env.pop(k, None)  # Delete key k, tolerating it being already gone
            else:
                env[k] = v
        yield
    finally:
        for k in to_delete:
            env.pop(k, None)  # Delete key k, tolerating it being already gone
        for k, v in to_restore.items():
            os.environ[k] = v


class ControlledTime:  # This will move to dcicutils -kmp 7-May-2020
    """
    This class can be used in mocking datetime.datetime for things that do certain time-related actions.
    Everytime datetime.now() is called, time increments by a known amount (default 1 second). So:

        start_time = datetime.datetime.now()

        def sleepy_function():
            time.sleep(10)

        dt = ControlledTime()
        with unittest.mock.patch("datetime.datetime", dt):
            with unittest.mock.patch("time.sleep", dt.sleep):
                t0 = datetime.datetime.now()
                sleepy_function()  # sleeps 10 seconds
                t1 = datetime.datetime.now()  # 1 more second increments
                assert (t1 - t0).total_seconds() == 11  # 11 virtual seconds have passed

        end_time = datetime.datetime.now()
        # In reality, whole test takes much less than one second...
        assert (end_time - start_time).total_seconds() < 0.5
    """

    # A randomly chosen but reproducible date 2010-07-01 12:00:00
    INITIAL_TIME = datetime.datetime(2010, 1, 1, 12, 0, 0)
    HMS_TIMEZONE = pytz.timezone("US/Eastern")
    _DATETIME = datetime.datetime

    def __init__(self, initial_time: datetime.datetime = INITIAL_TIME, tick_seconds: float = 1,
                 local_timezone: pytz.timezone = HMS_TIMEZONE):
        if not isinstance(initial_time, datetime.datetime):
            raise ValueError("Expected initial_time to be a datetime: %r" % initial_time)
        if initial_time.tzinfo is not None:
            raise ValueError("Expected initial_time to be a naive datetime (no timezone): %r" % initial_time)
        if not isinstance(tick_seconds, (int, float)):
            raise ValueError("Expected tick_seconds to be an int or a float: %r" % tick_seconds)

        self._initial_time = initial_time
        self._just_now = initial_time
        self._tick_timedelta = datetime.timedelta(seconds=tick_seconds)
        self._local_timezone = local_timezone

    def set_datetime(self, dt):
        """
        Sets the virtual clock time to the given value, which must be a regular datetime object.
        """
        if not isinstance(dt, datetime.datetime):
            raise ValueError("Expected a datetime: %r" % dt)
        if dt.tzinfo:
            raise ValueError("Expected a naive datetime (no timezone): %r" % dt)
        self._just_now = dt

    def reset_datetime(self):
        """
        Resets the virtual clock time to the original datetime that it had on creation.
        """
        self.set_datetime(self._initial_time)

    def just_now(self) -> datetime.datetime:
        """
        This is like .now() but it doesn't increment the virtual clock,
        it just tells you the previously known virtual clock time.
        """
        return self._just_now

    def now(self) -> datetime.datetime:
        """
        This advances time by one tick and returns the new time.

        To re-read the same time without advancing the clock, use .just_now() instead of .now().

        The tick duration is controlled by the tick_seconds initialization argument to ControlledTime.

        Note that neither .now() nor .utcnow() yields a time with a timezone,
        though the normal conversion operations are available on the times it does yield.
        It is assumed that .now() returns time in the Harvard Medical School timezone, US/Eastern
        unless a different local_timezone was specified when creating the ControlledTime.
        """
        self._just_now += self._tick_timedelta
        return self._just_now

    def utcnow(self) -> datetime.datetime:
        """
        This tells you what the virtual clock time would be in UTC.
        This works by adjusting for the difference in hours between the local timezone and UTC.

        Note that neither .now() nor .utcnow() yields a time with a timezone,
        though the normal conversion operations are available on the times it does yield.
        It is assumed that .now() returns time in the Harvard Medical School timezone, US/Eastern
        unless a different local_timezone was specified when creating the ControlledTime.
        """
        now = self.now()
        return self._local_timezone.localize(now).astimezone(pytz.UTC).replace(tzinfo=None)

    def sleep(self, secs: float):
        """
        This simulates sleep by advancing the virtual clock time by the indicated number of seconds.
        """

        self._just_now += datetime.timedelta(seconds=secs)


def notice_pytest_fixtures(*fixtures):
    """
    This declares its arguments to be pytest fixtures in use by surrounding code.

    This useful for assuring tools like flake8 and PyCharm that the arguments it is given are not being
    ignored but instead may just have usage patterns that don't make uses apparent.

    For example, in a file test_file.py, we might see uses of my_fixture and my_autoused_fixture
    that wrongly appear both globally unused AND also locally unused in the test_something function.

      from module_a import foo, bar
      from module_b import mock_application  # <-- sample fixture, to be used explicitly
      from module_c import mock_database     # <-- sample fixture to be used implicitly

      def test_something(mock_application):  # <-- sample of fixture used explicitly
          assert foo() = bar()

    In both cases, the apparently unused variables may cause flake8, PyCharm, and other
    'lint' programs to complain that some repair action is needed, such as removing the unused
    variables, even though such an action might break code. Using this declaration will
    guard against that, while providing useful documentation for the code:

      from module_a import foo, bar
      from module_b import mock_application
      from module_c import mock_database
      from dcicutils.qa_utils import notice_pytest_fixtures

      notice_pytest_fixtures(application, database_session)

      def test_something(mock_application):
          notice_pytest_fixtures(mock_application)  # <-- protects bound variable from seeming unused
          assert foo() = bar()
    """
    ignored(fixtures)  # we don't use the given fixtures, but now the tools will think we do


class Occasionally:
    """
    This class is useful for testing flakey things.

    Occasionally(function) returns a function that works the first time but fails every other time.
    Occasionally(function, success_frequency=N) returns a function that fails N-1 times, then works.
    Occasionally(function, failure_frequency=N) returns a function that works N-1 times, then fails.

    The counting phase can be reset by calling .reset() on the function.
    """
    # More examples of this can be seen in the tests for misc_utils.RetryManager

    DEFAULT_ERROR_CLASS = Exception

    DEFAULT_ERROR_MESSAGE = "Oops. Occasionally this fails."

    def __init__(self, function, failure_frequency=None, success_frequency=None,
                 error_class=None, error_message=None):
        if not failure_frequency and not success_frequency:
            # Arbitrary. Setting success_frequency would amount to the same, but on other 'phase'.
            # This will succeed once, then fail.
            failure_frequency = 2
        self.frequency = failure_frequency or success_frequency
        self.frequently_fails = True if success_frequency else False
        self.count = 0
        self.function = function
        self.error_class = error_class or self.DEFAULT_ERROR_CLASS
        self.error_message = error_message or self.DEFAULT_ERROR_MESSAGE

    def reset(self):
        self.count = 0

    def __call__(self, *args, **kwargs):
        self.count = (self.count + 1) % self.frequency
        if (self.count == 0 and not self.frequently_fails) or (self.count != 0 and self.frequently_fails):
            raise self.error_class(self.error_message)
        else:
            return self.function(*args, **kwargs)

    @property
    def __name__(self):
        return "{}.occassionally".format(self.function.__name__)


class RetryManager(Retry):
    """
    This class provides a method that is not thread-safe but is usable in unit testing to locally bind retry options
    for a function declared elsewhere using the Retry.allow_retries(...) decorator.

    NOTE: this does NOT use module names of functions to avoid collisions (though you can override that
    with a name_key). For example:

        with RetryManager.retry_options("foo", wait_seconds=4):
            ...code where foo will wait 4 seconds before retrying...

    """

    @classmethod
    @contextlib.contextmanager
    def retry_options(cls, name_key, retries_allowed=None, wait_seconds=None,
                      wait_increment=None, wait_multiplier=None):
        if not isinstance(name_key, str):
            raise ValueError("The required 'name_key' argument to the RetryManager.retry_options context manager"
                             " must be a string: %r" % name_key)
        function_profile = cls._RETRY_OPTIONS_CATALOG.get(name_key)
        if not function_profile:
            raise ValueError("The 'name_key' argument to RetryManager.retry_options"
                             " did not name a registered function: %r" % name_key)
        assert isinstance(function_profile, cls.RetryOptions)
        options = {}
        if retries_allowed is not None:
            options['retries_allowed'] = retries_allowed
        if wait_seconds is not None:
            options['wait_seconds'] = wait_seconds
        if wait_increment is not None:
            options['wait_increment'] = wait_increment
        if wait_multiplier is not None:
            options['wait_multiplier'] = wait_multiplier
        if wait_increment is not None or wait_multiplier is not None:
            options['wait_adjustor'] = function_profile.make_wait_adjustor(wait_increment=wait_increment,
                                                                           wait_multiplier=wait_multiplier)
        with local_attrs(function_profile, **options):
            yield


FILE_SYSTEM_VERBOSE = True


class MockFileWriter:

    def __init__(self, file_system, file, binary=False, encoding='utf-8'):
        self.file_system = file_system
        self.file = file
        self.encoding = encoding
        self.stream = io.BytesIO() if binary else io.StringIO()

    def __enter__(self):
        return self.stream

    def __exit__(self, exc_type, exc_val, exc_tb):
        content = self.stream.getvalue()
        if FILE_SYSTEM_VERBOSE:
            print("Writing %r to %s." % (content, self.file))
        self.file_system.files[self.file] = content if isinstance(content, bytes) else content.encode(self.encoding)


class MockFileSystem:
    """Extremely low-tech mock file system."""

    def __init__(self, files=None, default_encoding='utf-8'):
        self.default_encoding = default_encoding
        self.files = {filename: content.encode(default_encoding) for filename, content in (files or {}).items()}

    def exists(self, file):
        return bool(self.files.get(file))

    def remove(self, file):
        if not self.files.pop(file, None):
            raise FileNotFoundError("No such file or directory: %s" % file)

    def open(self, file, mode='r'):
        if FILE_SYSTEM_VERBOSE:
            print("Opening %r in mode %r." % (file, mode))
        if mode == 'w':
            return self._open_for_write(file_system=self, file=file, binary=False)
        elif mode == 'wb':
            return self._open_for_write(file_system=self, file=file, binary=True)
        elif mode == 'r':
            return self._open_for_read(file, binary=False)
        elif mode == 'rb':
            return self._open_for_read(file, binary=True)
        else:
            raise AssertionError("Mocked io.open doesn't handle mode=%r." % mode)

    def _open_for_read(self, file, binary=False, encoding=None):
        content = self.files.get(file)
        if content is None:
            raise FileNotFoundError("No such file or directory: %s" % file)
        if FILE_SYSTEM_VERBOSE:
            print("Read %r to %s." % (content, file))
        return io.BytesIO(content) if binary else io.StringIO(content.decode(encoding or self.default_encoding))

    def _open_for_write(self, file_system, file, binary=False, encoding=None):
        return MockFileWriter(file_system=file_system, file=file, binary=binary,
                              encoding=encoding or self.default_encoding)


class NotReallyRandom:

    def __init__(self):
        self.counter = 0

    def _random_int(self, n):
        """Returns an integer between 0 and n, upper-exclusive, not one of the published 'random' operations."""
        result = self.counter % n
        self.counter += 1
        return result

    def randint(self, a, b):
        """Returns a number between a and b, inclusive at both ends, though not especially randomly."""
        assert isinstance(a, int) and isinstance(b, int) and a < b, "Arguments must be two strictly ascending ints."
        rangesize = int(abs(b-a))+1
        return a + self._random_int(rangesize)

    def choice(self, things):
        return things[self._random_int(len(things))]



'''Utilities related to ElasticBeanstalk deployment and management.
This includes, but is not limited to: ES, s3, RDS, Auth0, and Foursight.
'''
from __future__ import print_function
import subprocess
import logging
import boto3
import os
import json
import requests
import sys
import time
from datetime import datetime
from . import ff_utils
from botocore.exceptions import ClientError
from .misc_utils import PRINT
from .env_utils import (
    is_fourfront_env, is_cgap_env, is_stg_or_prd_env, public_url_mappings,
    blue_green_mirror_env, get_standard_mirror_env,
)

logging.basicConfig()
logger = logging.getLogger('logger')
logger.setLevel(logging.INFO)

# In Python 2, the safe 'input' function was called 'raw_input'.  Also in Python 2, there was a function
# named 'input' that did eval(raw_input(...)).  Python 3 made an incompatible change, renaming 'raw_input'
# to 'input', and it no longer has a function that does an unsafe eval.  When we supported both Python 2 & 3,
# use a 'try' expression to sort things out and call the safe function 'use_input' to avoid confusion.
# But PyCharm found that 'try' expression confusing, so now that we are Python 3 only, we're phasing that
# out. For a time, we'll retain the transitional naming, though, along with an affirmative error check, so
# we don't open any security holes. We can remove this naming and check once we're we're only using Python 3.
# -kmp 27-Mar-2020

_python_major_version = sys.version_info[0]
if _python_major_version < 3:
    raise EnvironmentError("The 'dcicutils.beanstalk_utils' package only works in Python 3.")
use_input = input  # In Python 3, this does 'safe' input reading.


FOURSIGHT_URL = 'https://foursight.4dnucleome.org/'

# FF_MAGIC_CNAME corresponds to data.4dnucleome.org
FF_MAGIC_CNAME = 'fourfront-green.us-east-1.elasticbeanstalk.com'
# CGAP_MAGIC_CNAME corresponds to cgap.hms.harvard.edu
CGAP_MAGIC_CNAME = 'fourfront-cgap.9wzadzju3p.us-east-1.elasticbeanstalk.com'
# The legacy name MAGIC_CNAME is deprecated (retained for backward compatibility until a major release boundary).
MAGIC_CNAME = FF_MAGIC_CNAME

# FF_GOLDEN_DB is the database behind data.4dnucleome.org (and shared by staging.4dnucleome.org)
FF_GOLDEN_DB = 'fourfront-production.co3gwj7b7tpq.us-east-1.rds.amazonaws.com'
# CGAP_GOLDEN_DB is the database behind cgap.hms.harvard.edu
CGAP_GOLDEN_DB = 'fourfront-cgap.co3gwj7b7tpq.us-east-1.rds.amazonaws.com'
# The name GOLDEN_DB is deprecated (retained for backward compatibility until a major release boundary).
# Although not visibly used in this repository, this variable is imported by Torb.
GOLDEN_DB = FF_GOLDEN_DB

REGION = 'us-east-1'


class WaitingForBoto3(Exception):
    pass


def delete_db(db_identifier, take_snapshot=True, allow_delete_prod=False):
    """
    Given db_identifier, delete an RDS instance. If take_snapshot is true,
    will create a final snapshot named "<db_identifier>-final-<yyyy-mm-dd>".

    Args:
        db_identifier (str): name of RDS instance
        take_snapshot (bool): If True, take a final snapshot before deleting
        allow_delete_prod (bool): Must be True to allow deletion of 'webprod' DB

    Returns:
        dict: boto3 response from delete_db_instance
    """
    # safety. Do not allow accidental programmatic deletion of webprod DB
    if 'prod' in db_identifier and not allow_delete_prod:
        raise Exception('Must set allow_delete_prod to True to delete RDS instance' % db_identifier)
    client = boto3.client('rds')
    timestamp = datetime.strftime(datetime.utcnow(), "%Y-%m-%d")
    if take_snapshot:
        try:
            resp = client.delete_db_instance(
                DBInstanceIdentifier=db_identifier,
                SkipFinalSnapshot=False,
                FinalDBSnapshotIdentifier=db_identifier + "-final-" + timestamp
            )
        except:  # noqa: E722
            # Snapshot cannot be made. Likely a date conflict
            resp = client.delete_db_instance(
                DBInstanceIdentifier=db_identifier,
                SkipFinalSnapshot=True,
            )
    else:
        resp = client.delete_db_instance(
            DBInstanceIdentifier=db_identifier,
            SkipFinalSnapshot=True,
        )
    PRINT(resp)
    return resp


def get_es_from_bs_config(env):
    """
    Given an ElasticBeanstalk environment name, get the corresponding
    Elasticsearch url from the EB configurationock

    Args:
        env (str): ElasticBeanstalk environment name

    Returns:
        str: Elasticsearch url without port info
    """
    bs_env = get_bs_env(env)
    for item in bs_env:
        if item.startswith('ES_URL'):
            return item.split('=')[1].strip(':80')


def is_indexing_finished(env, prev_version=None, travis_build_id=None):
    """
    Checker function used with torb waitfor lambda; output must be standarized.
    Check to see if indexing of a Fourfront environment corresponding to given
    ElasticBeanstalk environment is finished by looking at the /counts page.

    Args:
        env (str): ElasticBeanstalk environment name
        prev_version (str): optional EB version of the previous configuration
        travis_build_id (int): optional ID for a Travis build

    Returns:
        bool, list: True if done, results from /counts page

    Raises:
        Exception: if Travis done and bad EB environment VersionLabel
        WaitingForBoto3: on a retryable waitfor condition
    """
    # is_beanstalk_ready will raise WaitingForBoto3 if not ready
    is_beanstalk_ready(env)
    bs_url = get_beanstalk_real_url(env)
    if not bs_url.endswith('/'):
        bs_url += "/"

    # retry if the beanstalk version has not updated from previous version,
    # unless the travis build has failed (in which case it will never update).
    # If failed, let is_indexing continue as usual so the deployment will
    # complete, despite being on the wrong EB application version
    if prev_version and beanstalk_info(env).get('VersionLabel') == prev_version:
        if travis_build_id:
            try:
                trav_done, trav_details = is_travis_finished(travis_build_id)
            except Exception as exc:
                # if the build failed, let the indexing check continue
                if 'Build Failed' in str(exc):
                    logger.info("EB version has not updated from %s."
                                "Associated travis build %s has failed."
                                % (prev_version, travis_build_id))
                else:
                    raise WaitingForBoto3("EB version has not updated from %s. "
                                          "Encountered error when getting build"
                                          " %s from Travis. Error: %s"
                                          % (prev_version, travis_build_id, exc))
            else:
                # Travis build is running/has passed
                if trav_done is True:
                    logger.info("EB version has not updated from %s."
                                "Associated travis build %s has finished."
                                % (prev_version, travis_build_id))
                else:
                    raise WaitingForBoto3("EB version has not updated from %s."
                                          "Associated travis build is %s and "
                                          "has not yet finished. Details: %s"
                                          % (prev_version, travis_build_id, trav_details))
        else:
            # no build ID provided; must retry on not updated version
            raise WaitingForBoto3("EB version has not updated from %s"
                                  % prev_version)

    # check counts from the portal to determine indexing state
    try:
        counts_res = ff_utils.authorized_request(bs_url + 'counts?format=json',
                                                 ff_env=env)
        totals = counts_res.json().get('db_es_total').split()

        # example value of split totals: ["DB:", "74048", "ES:", "74048"]
        db_total = totals[1]
        es_total = totals[3]

        if int(db_total) > int(es_total):
            is_ready = False
        else:
            is_ready = True
    except Exception as exc:
        logger.info('Error on is_indexing_finished: %s' % exc)
        is_ready = False
        totals = []
    return is_ready, totals


def swap_cname(src, dest):
    """
    Swap the CNAMEs of two ElasticBeanstalk (EB) environments, given by
    src and dest. Will restart the app servers after swapping.
    Should be used for swapping production/staging environments.

    Args:
        src (str): EB environment name of production
        dest (str): EB environment name of staging

    Returns:
        None
    """
    client = boto3.client('elasticbeanstalk', region_name=REGION)
    PRINT("Swapping CNAMEs %s and %s..." % (src, dest))
    client.swap_environment_cnames(SourceEnvironmentName=src,
                                   DestinationEnvironmentName=dest)
    PRINT("Giving CNAMEs 10 seconds to update...")
    time.sleep(10)
    PRINT("Restarting app servers for %s and %s..." % (src, dest))
    client.restart_app_server(EnvironmentName=src)
    client.restart_app_server(EnvironmentName=dest)


def _compute_prd_env_for_project(project):
    '''
    Determines which ElasticBeanstalk environment is currently hosting
    data.4dnucleome.org. Requires IAM permissions for EB!

    Returns:
        str: EB environment name hosting data.4dnucleome
    '''
    magic_cname = CGAP_MAGIC_CNAME if project == 'cgap' else FF_MAGIC_CNAME
    client = boto3.client('elasticbeanstalk', region_name=REGION)
    res = describe_beanstalk_environments(client, ApplicationName="4dn-web")
    for env in res['Environments']:
        if env.get('CNAME') == magic_cname:
            # we found data
            return env.get('EnvironmentName')


def compute_ff_prd_env():
    """Returns the name of the current Fourfront production environment."""
    return _compute_prd_env_for_project('fourfront')


def compute_ff_stg_env():
    """Returns the name of the current Fourfront staging environment."""
    return get_standard_mirror_env(compute_ff_prd_env())


whodaman = compute_ff_prd_env  # This naming is deprecated but retained for compatibility.


def compute_cgap_prd_env():
    """Returns the name of the current CGAP production environment."""
    return _compute_prd_env_for_project('cgap')


def compute_cgap_stg_env():
    """Returns the name of the current CGAP staging environment, or None if there is none."""
    return get_standard_mirror_env(compute_cgap_prd_env())


def compute_prd_env_for_env(envname):
    """Given an environment, returns the name of the prod environment for its owning project."""
    if is_cgap_env(envname):
        return compute_cgap_prd_env()
    elif is_fourfront_env(envname):
        return compute_ff_prd_env()
    else:
        raise ValueError("Unknown environment: %s" % envname)


def beanstalk_info(env):
    """
    Describe a ElasticBeanstalk environment given an environment name

    Args:
        env (str): ElasticBeanstalk environment name

    Returns:
        dict: Environments result from describe_beanstalk_environments
    """
    client = boto3.client('elasticbeanstalk', region_name=REGION)
    res = describe_beanstalk_environments(client, EnvironmentNames=[env])
    return res['Environments'][0]


def get_beanstalk_real_url(env):
    """
    Return the real url for the elasticbeanstalk with given environment name.
    Name can be 'cgap', 'data', 'staging', or an actual environment.

    Args:
        env (str): ElasticBeanstalk environment name

    Returns:
        str: url of the ElasticBeanstalk environment
    """
    url = ''
    urls = public_url_mappings(env)

    if env in urls:  # Special case handling of 'cgap', 'data', or 'staging' as an argument.
        return urls[env]

    if is_stg_or_prd_env(env):
        # What counts as staging/prod depends on whether we're in the CGAP or Fourfront space.
        data_env = compute_cgap_prd_env() if is_cgap_env(env) else compute_ff_prd_env()
        # There is only one production environment. Everything else is staging, but everything
        # else is not staging.4dnucleome.org. Only one is that.
        if env == data_env:
            return urls['data']
        elif env == blue_green_mirror_env(data_env):
            # Mirror env might be None, in which case this clause will not be entered
            return urls['staging']

    bs_info = beanstalk_info(env)
    url = "http://" + bs_info['CNAME']
    return url


def is_beanstalk_ready(env):
    """
    Checker function used with torb waitfor lambda; output must be standarized.
    Check to see if a ElasticBeanstalk environment status is "Ready"

    Args:
        env (str): ElasticBeanstalk environment name

    Returns:
        bool, str: True if done, ElasticBeanstalk url

    Raises:
        WaitingForBoto3: if EB environment status != "Ready"
    """
    client = boto3.client('elasticbeanstalk', region_name=REGION)
    res = describe_beanstalk_environments(client, EnvironmentNames=[env])

    status = res['Environments'][0]['Status']
    if status != 'Ready':
        raise WaitingForBoto3("Beanstalk environment status is %s" % status)

    return True, 'http://' + res['Environments'][0].get('CNAME')


def describe_beanstalk_environments(client, **kwargs):
    """
    Generic function for retrying client.describe_environments to avoid
    AWS throttling errors. Passes all given kwargs to describe_environments

    Args:
        client (botocore.client.ElasticBeanstalk): boto3 client

    Returns:
        dict: response from client.describe_environments

    Raises:
        Exception: if a non-ClientError exception is encountered during
            describe_environments or cannot complete within retry framework
    """
    env_info = kwargs.get('EnvironmentNames', kwargs.get('ApplicationName', 'Unknown environment'))
    for retry in [1, 1, 1, 1, 2, 2, 2, 4, 4, 6, 8, 10, 12, 14, 16, 18, 20]:
        try:
            res = client.describe_environments(**kwargs)
        except ClientError as e:
            PRINT('Client exception encountered while getting BS info for %s. Error: %s' % (env_info, str(e)))
            time.sleep(retry)
        except Exception as e:
            PRINT('Unhandled exception encountered while getting BS info for %s. Error: %s' % (env_info, str(e)))
            raise e
        else:
            return res
    raise Exception('Could not describe Beanstalk environments due ClientErrors, likely throttled connections.')


def is_snapshot_ready(snapshot_name):
    """
    Checker function used with torb waitfor lambda; output must be standarized.
    Check to see if an RDS snapshot with given name is available

    Args:
        snapshot_name (str): RDS snapshot name

    Returns:
        bool, str: True if done, identifier of snapshot
    """
    client = boto3.client('rds', region_name=REGION)
    resp = client.describe_db_snapshots(DBSnapshotIdentifier=snapshot_name)
    db_status = resp['DBSnapshots'][0]['Status']
    is_ready = db_status.lower() == 'available'
    return is_ready, resp['DBSnapshots'][0]['DBSnapshotIdentifier']


def is_es_ready(es_name):
    """
    Checker function used with torb waitfor lambda; output must be standarized.
    Check to see if an ES instance is ready and has an endpoint

    Args:
        es_name (str): ES instance name

    Returns:
        bool, str: True if done, ES url
    """
    es = boto3.client('es', region_name=REGION)
    describe_resp = es.describe_elasticsearch_domain(DomainName=es_name)
    endpoint = describe_resp['DomainStatus'].get('Endpoint', None)
    is_ready = True
    if endpoint is None:
        is_ready = False
    else:
        endpoint = endpoint + ":80"
    return is_ready, endpoint


def is_db_ready(db_identifier):
    """
    Checker function used with torb waitfor lambda; output must be standarized.
    Check to see if an RDS instance with given name is ready

    Args:
        db_identifier (str): RDS instance identifier

    Returns:
        bool, str: True if done, RDS address
    """
    client = boto3.client('rds', region_name=REGION)
    is_ready = False
    resp = client.describe_db_instances(DBInstanceIdentifier=db_identifier)
    details = resp
    endpoint = resp['DBInstances'][0].get('Endpoint')
    if endpoint and endpoint.get('Address'):
        details = endpoint['Address']
        is_ready = True

    return is_ready, details


def create_db_snapshot(db_identifier, snapshot_name):
    """
    Given an RDS instance indentifier, create a snapshot using the given name.
    If a snapshot with given name already exists, attempt to delete and return
    "Deleting". Otherwise, return snapshot ARN.

    Args:
        db_identifier (str): RDS instance identifier
        snapshot_name (str): identifier/ARN of RDS snapshot to create

    Returns:
        str: resource ARN if successful, otherwise "Deleting"
    """
    client = boto3.client('rds', region_name=REGION)
    try:
        response = client.create_db_snapshot(
             DBSnapshotIdentifier=snapshot_name,
             DBInstanceIdentifier=db_identifier)
    except ClientError:
        # probably the guy already exists
        try:
            client.delete_db_snapshot(DBSnapshotIdentifier=snapshot_name)
        except ClientError:
            pass
        return "Deleting"

    return response


def create_db_from_snapshot(db_identifier, snapshot_name, delete_db=True):
    """
    Given an RDS instance indentifier and a snapshot ARN/name, create an RDS
    instance from the snapshot. If an instance already exists with the given
    identifier and delete_db is True, attempt to delete and return "Deleting".
    Otherwise, return instance ARN.

    Args:
        db_identifier (str): RDS instance identifier
        snapshot_name (str): identifier/ARN of RDS snapshot to restore from

    Returns:
        str: resource ARN if successful, otherwise "Deleting"
    """
    client = boto3.client('rds', region_name=REGION)
    try:
        response = client.restore_db_instance_from_db_snapshot(
            DBInstanceIdentifier=db_identifier,
            DBSnapshotIdentifier=snapshot_name,
            DBInstanceClass='db.t2.medium',
            StorageType='gp2',
            PubliclyAccessible=False
        )
    except ClientError:
        # Something went wrong
        # Even if delete_db, never allow deletion of a db with 'webprod' in it
        if delete_db:
            # Drop target database with final snapshot
            try:
                delete_db(db_identifier, True)
            except ClientError:
                pass
            return "Deleting"
        else:
            return "Error"
    return response['DBInstance']['DBInstanceArn']


def is_travis_started(request_url):
    """
    Checker function used with torb waitfor lambda; output must be standarized.
    Check the requests url to see if a given build has stared and been issued
    a build id, which can in turn be used for is_travis_finished

    Args:
        request_url (str): Travis request url

    Returns:
        bool, dict: True if started, Travis response JSON

    Raises:
        Exception: if Travis key not in environ
    """
    if 'travis_key' not in os.environ:
        raise Exception('Must have travis_key environment variable defined')
    is_ready = False
    details = 'requested build has not started'
    headers = {'Content-Type': 'application/json',
               'Accept': 'application/json',
               'Travis-API-Version': '3',
               'User-Agent': 'tibanna/0.1.0',
               'Authorization': 'token %s' % os.environ['travis_key']}
    resp = requests.get(request_url, headers=headers)
    if resp.ok:
        logger.info("Travis request response (okay): %s" % resp.json())
        details = resp.json()
        if len(resp.json().get('builds', [])) == 1:
            is_ready = True
    return is_ready, details


def is_travis_finished(build_id):
    """
    Checker function used with torb waitfor lambda; output must be standarized.
    Check to see if a given travis build has passed

    Args:
        build_id (str): Travis build identifier

    Returns:
        bool, dict: True if done, Travis response JSON

    Raises:
        Exception: if the Travis build failed or Travis key not in environ
    """
    if 'travis_key' not in os.environ:
        raise Exception('Must have travis_key environment variable defined')
    is_ready = False
    details = 'build not done or not found'
    headers = {'Content-Type': 'application/json',
               'Accept': 'application/json',
               'Travis-API-Version': '3',
               'User-Agent': 'tibanna/0.1.0',
               'Authorization': 'token %s' % os.environ['travis_key']}

    url = 'https://api.travis-ci.org/build/%s' % build_id

    logger.info("Travis build url: %s" % url)
    resp = requests.get(url, headers=headers)
    logger.info("Travis build response: %s" % resp.text)
    state = resp.json()['state']
    if resp.ok and state in ['failed', 'errored']:
        raise Exception('Build Failed')
    elif resp.ok and state == 'passed':
        is_ready = True
        details = resp.json()
    return is_ready, details


def make_envvar_option(name, value):
    return {'Namespace': 'aws:elasticbeanstalk:application:environment',
            'OptionName': name,
            'Value': value}


def get_bs_env(envname):
    """
    Given an ElasticBeanstalk environment name, get the env variables from that
    environment and return them. Returned variables are in form: <name>=<value>

    Args:
        envname (str): name of ElasticBeanstalk environment

    Returns:
        list: of environment variables in <name>=<value> form
    """
    client = boto3.client('elasticbeanstalk', region_name=REGION)

    data = client.describe_configuration_settings(EnvironmentName=envname,
                                                  ApplicationName='4dn-web')
    options = data['ConfigurationSettings'][0]['OptionSettings']
    env_vars = [option['Value'] for option in options
                if option['OptionName'] == 'EnvironmentVariables'][0]
    return env_vars.split(',')


def update_bs_config(envname, template=None, keep_env_vars=False,
                     env_override=None):
    """
    Update the configuration for an existing ElasticBeanstalk environment.
    Requires the environment name. Can optionally specify a configuration
    template, as well as keep all environment variables from the existing
    environment with optional variable overrides.

    Args:
        envname (str): name of the EB environment
        template (str): configuration template to use. Default None
        keep_env_vars (bool): if True, keep existing env vars. Default False
        env_override (dict): if provided, overwrite existing env vars using the
            given key/values. Must use keep_env_vars to work. Default None

    Returns:
        dict: update_environment response
    """
    if template is None and not keep_env_vars:
        # nothing to update
        logger.info("update_bs_config: nothing to update for env %s!" % envname)
        return None
    client = boto3.client('elasticbeanstalk', region_name=REGION)
    options = []  # used to hold env vars
    if keep_env_vars:
        env_vars = get_bs_env(envname)
        for var in env_vars:
            key, value = var.split('=')
            if env_override and env_override.get(key):
                options.append(make_envvar_option(key, env_override[key]))
            else:
                options.append(make_envvar_option(key, value))

    # update template and/or env var options
    if options and template:
        return client.update_environment(EnvironmentName=envname,
                                         TemplateName=template,
                                         OptionSettings=options)
    elif template:
        return client.update_environment(EnvironmentName=envname,
                                         TemplateName=template)
    else:
        return client.update_environment(EnvironmentName=envname,
                                         OptionSettings=options)


def create_bs(envname, load_prod, db_endpoint, es_url, for_indexing=False):
    """
    XXX: Will not work currently, do NOT use on production

    Create a beanstalk environment given an envname. Use customized options,
    configuration template, and environment variables. If adding new env vars,
    make sure to overwrite them here.
    If the environment already exists, will update it with `update_bs_config`

    Args:
        envname (str): ElasticBeanstalk (EB) enviroment name
        load_prod (bool): sets the LOAD_FUNCTION EB env var
        db_endpoint (str): sets the RDS_HOSTNAME EB env var
        es_url (str): sets the ES_URL EB env var
        for_indexing (bool): If True, use 'fourfront-indexing' config template

    Returns:
        dict: boto3 res from create_environment/update_environment
    """
    # TODO (C4-280): Reconsider this and other functionality.
    if is_stg_or_prd_env(envname):
        raise RuntimeError("beanstalk_utils.create_bs is not approved for production use.")

    client = boto3.client('elasticbeanstalk', region_name=REGION)

    # determine the configuration template for Elasticbeanstalk
    template = 'fourfront-base'
    if for_indexing:
        template = 'fourfront-indexing'
    load_value = 'load_test_data'
    if load_prod:
        load_value = 'load_prod_data'

    options = [
        make_envvar_option('RDS_HOSTNAME', db_endpoint),
        make_envvar_option('ENV_NAME', envname),
        make_envvar_option('ES_URL', es_url),
        make_envvar_option('LOAD_FUNCTION', load_value)
    ]

    # logic for mirrorEsEnv, which is used to coordinate elasticsearch
    # changes between fourfront data and staging
    if 'fourfront-webprod' in envname:
        # TODO: This code is obsolete and needs to be upgraded. For now, the use of this on production is disabled.
        other_env = 'fourfront-webprod2' if envname == 'fourfront-webprod' else 'fourfront-webprod'
        mirror_es = get_es_build_status(other_env, max_tries=3)
        if mirror_es:
            options.append(make_envvar_option('mirrorEnvEs', mirror_es))

    try:
        res = client.create_environment(
            ApplicationName='4dn-web',
            EnvironmentName=envname,
            TemplateName=template,
            OptionSettings=options,
        )
    except ClientError:
        # environment already exists update it with given template
        # parse out current env variables and override existing values
        env_vars = {opt['OptionName']: opt['Value'] for opt in options}
        res = update_bs_config(envname, template=template, keep_env_vars=True,
                               env_override=env_vars)
    return res


# location of environment variables on elasticbeanstalk
BEANSTALK_ENV_PATH = "/opt/python/current/env"


def source_beanstalk_env_vars(config_file=BEANSTALK_ENV_PATH):
    """
    set environment variables if we are on Elastic Beanstalk
    AWS_ACCESS_KEY_ID is indicative of whether or not env vars are sourced

    Args:
        config_file (str): filepath to load env vars from
    """
    if os.path.exists(config_file) and not os.environ.get("AWS_ACCESS_KEY_ID"):
        command = ['bash', '-c', 'source ' + config_file + ' && env']
        proc = subprocess.Popen(command, stdout=subprocess.PIPE, universal_newlines=True)
        for line in proc.stdout:
            key, _, value = line.partition("=")
            os.environ[key] = value[:-1]
        proc.communicate()


def log_to_foursight(event, lambda_name='', overrides=None):
    """
    Use Foursight as a logging tool within in a lambda function by doing a PUT
    to /api/checks. Requires that the event has "_foursight" key, which is a
    subobject with the following:
    fields:
        "check": required, in form "<fs environment>/<check name>"
        "log_desc": will set "summary" and "description" if those are missing
        "full_output": optional. If not provided, use to provide info on lambda
        "brief_output": optional
        "summary": optional. If not provided, use "log_desc" value
        "description": optional. If not provided, use "log_desc" value
        "status": optional. If not provided, use "WARN"
    Can also optionally provide an dictionary to overrides param, which will
    update the event["_foursight"]

    Args:
        event (dict): Event input, most likely from a lambda with a workflow
        lambda_name (str): Name of the lambda that is calling this
        overrides (dict): Optionally override event['_foursight'] with this

    Returns:
        Response object from foursight

    Raises:
        Exception: if cannot get body from Foursight response
    """
    fs = event.get('_foursight')
    if fs and fs.get('check'):
        if overrides is not None and isinstance(overrides, dict):
            fs.update(overrides)
        # handles these fields. set full_output as a special case
        full_output = fs.get('full_output', '%s started to run' % lambda_name)
        brief_output = fs.get('brief_output')
        summary = fs.get('summary', fs.get('log_desc'))
        description = fs.get('description', fs.get('log_desc'))
        status = fs.get('status', 'WARN')

        data = {'status': status, 'summary': summary, 'description': description,
                'full_output': full_output, 'brief_output': brief_output}
        fs_auth = os.environ.get('FS_AUTH')
        headers = {'content-type': "application/json", 'Authorization': fs_auth}
        # fs['check'] should be in form: "<fs environment>/<check name>"
        url = FOURSIGHT_URL + 'checks/' + fs['check']
        res = requests.put(url, data=json.dumps(data), headers=headers)
        PRINT('Foursight response from %s: %s' % (url, res.text))
        try:
            return res.json()
        except Exception as exc:
            raise Exception('Error putting FS check to %s with body %s. '
                            'Exception: %s. Response text: %s'
                            % (url, data, exc, res.text))


def create_foursight_auto(dest_env):
    """
    Call `create_foursight` to create a Foursight environment based off a
    just a ElasticBeanstalk environment name. Determines a number of fields
    needed for the environment creation automatically. Also causes initial
    checks to be run on the new FS environment.

    Args:
        dest_env (str): ElasticBeanstalk environment name

    Returns:
        dict: response from Foursight PUT /api/environments
    """
    fs = {'dest_env': dest_env}

    # automatically determine info for FS environ creation
    fs['bs_url'] = get_beanstalk_real_url(dest_env)
    fs['fs_url'] = get_foursight_env(dest_env, fs['bs_url'])
    fs['es_url'] = get_es_from_bs_config(dest_env)

    fs['foursight'] = create_foursight(**fs)
    if fs['foursight'].get('initial_checks'):
        del fs['foursight']['initial_checks']

    return fs


def get_foursight_env(dest_env, bs_url=None):
    """
    Get a Foursight environment name corresponding the given ElasticBeanstalk
    environment name, with optionally providing the EB url for must robustness

    Args:
        dest_env (str): ElasticBeanstalk environment name
        bs_url (str): optional url for the ElasticBeanstalk instance

    Returns:
        str: Foursight environment name
    """
    if not bs_url:
        bs_url = get_beanstalk_real_url(dest_env)
    env = dest_env
    if 'data.4dnucleome.org' in bs_url:
        env = 'data'
    elif 'staging.4dnucleome.org' in bs_url:
        env = 'staging'
    return env


def create_foursight(dest_env, bs_url, es_url, fs_url=None):
    """
    Creates a new Foursight environment based off of dest_env. Since Foursight
    environments don't include "fourfront-" by convention, remove this if part
    of the env. Take some other options for settings on the env

    Note: this will cause all checks in all schedules to be run, to initialize
          environment.

    Args:
        dest_env (str): ElasticBeanstalk environment name
        bs_url (str): url of the ElasticBeanstalk for FS env
        es_url (str): url of the ElasticSearch for FS env
        fs_url (str): If provided, use to override dest-env based FS url

    Returns:
        dict: response from Foursight PUT to /api/environments

    Raises:
        Exception: if cannot get body from Foursight response
    """
    # we want some url like thing
    if not bs_url.startswith('http'):
        bs_url = 'http://' + bs_url
    if not bs_url.endswith("/"):
        bs_url += "/"

    es_url = es_url.rstrip(":80")
    if not es_url.startswith("http"):
        es_url = "https://" + es_url
    if not es_url.endswith("/"):
        es_url += "/"

    # environments on foursight don't include fourfront
    if not fs_url:
        fs_url = dest_env
    if fs_url.startswith('fourfront-'):
        fs_url = fs_url[len('fourfront-'):]

    foursight_url = FOURSIGHT_URL + 'environments/' + fs_url
    payload = {"fourfront": bs_url,  "es": es_url, "ff_env": dest_env}

    ff_auth = os.environ.get('FS_AUTH')
    headers = {'content-type': "application/json", 'Authorization': ff_auth}
    res = requests.put(foursight_url, data=json.dumps(payload), headers=headers)
    try:
        return res.json()
    except Exception as exc:
        raise Exception('Error creating FS environ to %s with body %s. '
                        'Exception: %s. Response text: %s'
                        % (foursight_url, payload, exc, res.text))


def create_new_es(new):
    """
    Create a new Elasticsearch domain with given name. See the
    args below for the settings used.

    TODO: do we want to add cognito and access policy setup here?
          I think not, since that's a lot of info to put out there...

    Args:
        new (str): Elasticsearch domain name

    Returns:
        dict: response from boto3 client
    """
    es = boto3.client('es', region_name=REGION)
    resp = es.create_elasticsearch_domain(
        DomainName=new,
        ElasticsearchVersion='5.3',
        ElasticsearchClusterConfig={
            'InstanceType': 'm4.large.elasticsearch',
            'InstanceCount': 2
        },
        EBSOptions={
            "EBSEnabled": True,
            "VolumeType": "gp2",
            "VolumeSize": 50
        }
    )
    PRINT('=== CREATED NEW ES DOMAIN %s ===' % new)
    PRINT('NO MASTER INSTANCES ARE USED!')
    PRINT('MAKE SURE TO UPDATE COGNITO AND ACCESS POLICY USING THE GUI!')
    PRINT(resp)

    return resp


def get_es_build_status(new, max_tries=None):
    """
    Check the build status of an Elasticsearch instance with given name.
    If max_tries is provided, only allow that many iterative checks to ES.
    Returns the ES endpoint plus port (80)

    Args:
        new (str): ES instance name
        max_tries (int): max number of times to check. Default None (no limit)

    Returns:
        str: ES endpoint plus port, or None if not found in max_tries
    """
    es = boto3.client('es', region_name=REGION)
    endpoint = None
    tries = 0
    while endpoint is None:
        describe_resp = es.describe_elasticsearch_domain(DomainName=new)
        endpoint = describe_resp['DomainStatus'].get('Endpoint')
        if max_tries is not None and tries >= max_tries:
            break
        if endpoint is None:
            PRINT(".")
            tries += 1
            time.sleep(10)

    # aws uses port 80 for es connection, lets be specific
    if endpoint:
        endpoint += ":80"
    PRINT('Found ES endpoint for %s: %s' % (new, endpoint))
    return endpoint


#########################################################################
# Functions meant to be used locally to clone or remove a beanstalk ENV #
# A lot of these functions use command line tools.                      #
# Sort of janky, but could end up being helpful someday ...             #
#########################################################################


def add_es(new, force_new=False):
    """
    Either gets information on an existing Elasticsearch instance
    or, if force_new is True, will create a new instance.

    Args:
        new (str): Fourfront EB environment name used for ES instance
        force_new (bool): if True, make a new ES. Default False

    Returns:
        str: AWS ARN of the ES instance
    """
    es = boto3.client('es', region_name=REGION)
    if force_new:
        # fallback is a new ES env to use if cannot create one with
        # `new` environment name
        fallback = new
        if new.endswith("-a"):
            fallback = fallback.replace("-a", "-b")
        elif new.endswith("-b"):
            fallback = fallback.replace("-b", "-a")
        else:
            fallback += "-a"
        try:
            resp = create_new_es(new)
        except:  # noqa: E722
            resp = create_new_es(fallback)
    else:
        try:
            resp = es.describe_elasticsearch_domain(DomainName=new)
        except ClientError:  # its not there
            resp = create_new_es(new)
    return resp['DomainStatus']['ARN']


def delete_es_domain(env_name):
    """
    Given an Elasticsearch domain name, delete the domain

    Args:
        env_name (str): Fourfront EB environment name used for ES instance

    Returns:
        None
    """
    # get the status of this bad boy
    es = boto3.client('es')
    try:
        res = es.delete_elasticsearch_domain(DomainName=env_name)
        PRINT(res)
    except:  # noqa: E722
        PRINT("es domain %s not found, skipping" % env_name)


def clone_bs_env_cli(old, new, load_prod, db_endpoint, es_url):
    """
    Use the eb command line client to clone an ElasticBeanstalk environment
    with some extra options.

    Args:
        old (str): existing EB environment name
        new (str): new EB environment name
        load_prod (bool): determins LOAD_FUNCTION EB env var
        db_endpoint (str): determines RDS_HOSTNAME EB env var
        es_url (str): determines ES_URL EB env var

    Returns:
        None
    """
    env = 'RDS_HOSTNAME=%s,ENV_NAME=%s,ES_URL=%s' % (db_endpoint, new, es_url)
    if load_prod is True:
        env += ",LOAD_FUNCTION=load_prod_data"
    subprocess.check_call(['eb', 'clone', old, '-n', new,
                           '--envvars', env,
                           '--exact', '--nohang'])


def delete_bs_env_cli(env_name):
    """
    Use the eb command line client to remove an ElasticBeanstalk environment
    with some extra options.

    Args:
        env_name (str): EB environment name

    Returns:
        None
    """
    subprocess.check_call(['eb', 'terminate', env_name, '-nh', '--force'])


def create_s3_buckets(new):
    """
    Given an ElasticBeanstalk env name, create the following s3 buckets that
    are standard for any of our EB environments.

    Args:
        new (str): EB environment name

    Returns:
        None
    """
    new_buckets = [
        'elasticbeanstalk-%s-blobs' % new,
        'elasticbeanstalk-%s-files' % new,
        'elasticbeanstalk-%s-wfoutput' % new,
        'elasticbeanstalk-%s-system' % new,
    ]
    s3 = boto3.client('s3', region_name=REGION)
    for bucket in new_buckets:
        s3.create_bucket(Bucket=bucket)
    PRINT('=== CREATED NEW S3 BUCKETS ===' % new)
    PRINT('MAKE SURE TO UPDATE CORS POLICY FOR -files AND -wfoutput BUCKETS!')


def delete_s3_buckets(env_name):
    """
    Given an ElasticBeanstalk env name, remove the following s3 buckets that
    are standard for any of our EB environments.

    Args:
        env_name (str): EB environment name

    Returns:
        None
    """
    buckets = [
        'elasticbeanstalk-%s-blobs' % env_name,
        'elasticbeanstalk-%s-files' % env_name,
        'elasticbeanstalk-%s-wfoutput' % env_name,
        'elasticbeanstalk-%s-system' % env_name,
    ]

    s3 = boto3.resource('s3')
    for bucket in buckets:
        PRINT("deleting content for " + bucket)
        try:
            s3.Bucket(bucket).objects.delete()
            s3.Bucket(bucket).delete()
        except:  # noqa: E722
            PRINT(bucket + " not found skipping...")


def snapshot_and_clone_db(db_identifier, snapshot_name):
    """
    Given a RDS instance identifier and snapshot name, will create a snapshot
    with that name and then spin up a new RDS instance named after the snapshot

    Args:
        db_identifier (str): original RDS identifier of DB to snapshot
        snapshot_name (str): identifier of snapshot AND new instance

    Returns:
        str: address of the new instance
    """
    client = boto3.client('rds', region_name=REGION)
    snap_res = create_db_snapshot(db_identifier, snapshot_name)
    if snap_res == 'Deleting':
        snap_res = client.create_db_snapshot(
            DBSnapshotIdentifier=snapshot_name,
            DBInstanceIdentifier=db_identifier
        )
    PRINT("Response from create db snapshot: %s" % snap_res)
    PRINT("Waiting for snapshot to create...")
    waiter = client.get_waiter('db_snapshot_completed')
    waiter.wait(DBSnapshotIdentifier=snapshot_name)
    PRINT("Done waiting, creating a new database with name %s" % snapshot_name)
    db_res = create_db_from_snapshot(snapshot_name, snapshot_name, False)
    if db_res == 'Error':
        raise Exception('Could not create DB %s; already exists' % snapshot_name)
    PRINT("Waiting for DB to be created from snapshot...")
    endpoint = ''
    while not endpoint:
        resp = client.describe_db_instances(DBInstanceIdentifier=snapshot_name)
        endpoint = resp['DBInstances'][0].get('Endpoint')
        if endpoint and endpoint.get('Address'):
            PRINT("We got an endpoint:", endpoint['Address'])
            return endpoint['Address']
        PRINT(".")
        time.sleep(10)


def add_to_auth0_client(new):
    """
    Given an ElasticBeanstalk env name, find the url and use it to update the
    callback URLs for Auth0

    Args:
        new (str): EB environment name

    Returns:
        None
    """
    client = boto3.client('elasticbeanstalk', region_name=REGION)
    url = None
    PRINT("Getting beanstalk URL for env %s..." % new)
    while url is None:
        env = describe_beanstalk_environments(client, EnvironmentNames=[new])
        url = env['Environments'][0].get('CNAME')
        if url is None:
            PRINT(".")
            time.sleep(10)
    auth0_client_update(url)


def remove_from_auth0_client(env_name):
    """
    Given an ElasticBeanstalk env name, find the url and remove it from the
    callback urls for Auth0

    Args:
        env_name (str): EB environment name

    Returns:
        None
    """
    client = boto3.client('elasticbeanstalk', region_name=REGION)
    url = None
    PRINT("Getting beanstalk URL for env %s..." % env_name)
    while url is None:
        env = client.describe_environments(EnvironmentNames=[env_name])
        url = env['Environments'][0].get('CNAME')
        if url is None:
            PRINT(".")
            time.sleep(10)
    auth0_client_remove(url)


def auth0_client_update(url):
    """
    Get a JWT for programmatic access to Auth0 using Client/Secret env vars.
    Then add the given `url` to the Auth0 callbacks list.

    Args:
        url (str): url to add to callbacks

    Returns:
        None
    """
    # generate a jwt to validate future requests
    client = os.environ["Auth0Client"]
    secret = os.environ["Auth0Secret"]
    payload = {"grant_type": "client_credentials",
               "client_id": client,
               "client_secret": secret,
               "audience": "https://hms-dbmi.auth0.com/api/v2/"}
    headers = {'content-type': "application/json"}
    res = requests.post("https://hms-dbmi.auth0.com/oauth/token",
                        data=json.dumps(payload),
                        headers=headers)
    jwt = res.json()['access_token']

    client_url = "https://hms-dbmi.auth0.com/api/v2/clients/%s" % client
    headers['authorization'] = 'Bearer %s' % jwt

    get_res = requests.get(client_url + '?fields=callbacks', headers=headers)
    callbacks = get_res.json()['callbacks']
    callbacks.append("http://" + url)
    client_data = {'callbacks': callbacks}

    update_res = requests.patch(client_url, data=json.dumps(client_data), headers=headers)
    PRINT('auth0 callback urls are: %s' % update_res.json().get('callbacks'))


def auth0_client_remove(url):
    """
    Get a JWT for programmatic access to Auth0 using Client/Secret env vars.
    Then use that to remove the given `url` from the Auth0 callbacks list.

    Args:
        url (str): url to remove from callbacks

    Returns:
        None
    """
    # generate a jwt to validate future requests
    client = os.environ["Auth0Client"]
    secret = os.environ["Auth0Secret"]
    payload = {"grant_type": "client_credentials",
               "client_id": client,
               "client_secret": secret,
               "audience": "https://hms-dbmi.auth0.com/api/v2/"}
    headers = {'content-type': "application/json"}
    res = requests.post("https://hms-dbmi.auth0.com/oauth/token",
                        data=json.dumps(payload),
                        headers=headers)
    jwt = res.json()['access_token']
    client_url = "https://hms-dbmi.auth0.com/api/v2/clients/%s" % client
    headers['authorization'] = 'Bearer %s' % jwt

    get_res = requests.get(client_url + '?fields=callbacks', headers=headers)
    callbacks = get_res.json()['callbacks']
    full_url = 'http://' + url
    try:
        idx = callbacks.index(full_url)
    except ValueError:
        PRINT(full_url + " Not in auth0 auth, doesn't need to be removed")
        return
    if idx:
        callbacks.pop(idx)
    client_data = {'callbacks': callbacks}

    update_res = requests.patch(client_url, data=json.dumps(client_data), headers=headers)
    PRINT('auth0 callback urls are: %s' % update_res.json().get('callbacks'))


def copy_s3_buckets(new, old):
    """
    Given a new ElasticBeanstalk environment name and existing "old" one,
    create the given buckets and copy contents from the corresponding
    existing ones

    Args:
        new (str): new EB environment name
        old (str): existing EB environment name

    Returns:
        None
    """
    # each env needs the following buckets
    new_buckets = [
        'elasticbeanstalk-%s-blobs' % new,
        'elasticbeanstalk-%s-files' % new,
        'elasticbeanstalk-%s-wfoutput' % new,
        'elasticbeanstalk-%s-system' % new,
    ]
    old_buckets = [
        'elasticbeanstalk-%s-blobs' % old,
        'elasticbeanstalk-%s-files' % old,
        'elasticbeanstalk-%s-wfoutput' % old,
    ]
    s3 = boto3.client('s3', region_name=REGION)
    for bucket in new_buckets:
        try:
            s3.create_bucket(Bucket=bucket)
        except:  # noqa: E722
            PRINT("bucket %s already created..." % bucket)

    # now copy them
    # aws s3 sync s3://mybucket s3://backup-mybucket
    # get rid of system bucket
    new_buckets.pop()
    for old, new in zip(old_buckets, new_buckets):
        oldb = "s3://%s" % old
        newb = "s3://%s" % new
        PRINT("copying data from old %s to new %s" % (oldb, newb))
        subprocess.call(['aws', 's3', 'sync', oldb, newb])


def clone_beanstalk_command_line(old, new, prod=False, copy_s3=False):
    """
    Maybe useful command to clone an existing ElasticBeanstalk environment to
    a new one. Will create an Elasticsearch instance, s3 buckets, clone the
    existing RDS of the environment, and optionally copy s3 contents.
    Also adds the new EB url to Auth0 callback urls.
    Should be run exclusively via command line, as it requires manual input
    and subprocess calls of AWS command line tools.

    Note:
        The eb cli tool sets up a configuration file in the directory of the
        project respository. As such, this command MUST be called from that
        directory. Will exit if not called from an eb initialized directory.

    Args:
        old (str): environment name of existing ElasticBeanstalk
        new (str): new ElasticBeanstalk environment name
        prod (bool): set to True if this is a prod environment. Default False
        copy_s3 (bool): set to True to copy s3 contents. Default False

    Returns:
        None
    """
    if 'Auth0Client' not in os.environ or "Auth0Secret" not in os.environ:
        PRINT('Must set Auth0Client and Auth0Secret env variables! Exiting...')
        return
    PRINT('### eb status (START)')
    eb_ret = subprocess.call(['eb', 'status'])
    if eb_ret != 0:
        PRINT('This command must be called from an eb initialized repo! Exiting...')
        return
    PRINT('### eb status (END)')
    name = use_input("This will create an environment named %s, cloned from %s."
                     " This includes s3, ES, RDS, and Auth0 callbacks. If you "
                     "are sure, type the new env name to confirm: " % (new, old))
    if str(name) != new:
        PRINT("Could not confirm env. Exiting...")
        return
    PRINT("### start build ES service")
    add_es(new)
    PRINT("### create the s3 buckets")
    create_s3_buckets(new)
    PRINT("### create snapshot and copy database")
    db_endpoint = snapshot_and_clone_db(old, new)
    PRINT("### waiting for ES service")
    es_endpoint = get_es_build_status(new)
    PRINT("### clone elasticbeanstalk envrionment")
    # TODO, can we pass in github commit id here?
    clone_bs_env_cli(old, new, prod, db_endpoint, es_endpoint)
    PRINT("### allow auth-0 requests")
    add_to_auth0_client(new)
    if copy_s3 is True:
        PRINT("### copy contents of s3")
        copy_s3_buckets(new, old)
    PRINT("### All done! It may take some time for the beanstalk env to finish"
          " initialization. You may want to deploy the most current FF branch.")


def delete_beanstalk_command_line(env):
    """
    Maybe useful command to delete an existing ElasticBeanstalk environment,
    including associated ES, s3, and RDS resources. Will also remove the
    associated callback url from Auth0.
    Should be run exclusively via command line, as it requires manual input
    and subprocess calls of AWS command line tools.

    Note:
        The eb cli tool sets up a configuration file in the directory of the
        project respository. As such, this command MUST be called from that
        directory. Will exit if not called from an eb initialized directory.

    Args:
        env (str): EB environment name to delete

    Returns:
        None
    """
    if 'Auth0Client' not in os.environ or "Auth0Secret" not in os.environ:
        PRINT('Must set Auth0Client and Auth0Secret env variables! Exiting...')
        return
    PRINT('### eb status (START)')
    eb_ret = subprocess.call(['eb', 'status'])
    if eb_ret != 0:
        PRINT('This command must be called from an eb initialized repo! Exiting...')
        return
    PRINT('### eb status (END)')
    name = use_input("This will totally blow away the environment, including s3,"
                     "ES, RDS, and Auth0 callbacks. If you are sure, type the "
                     "env name to confirm: ")
    if str(name) != env:
        PRINT("Could not confirm env. Exiting...")
        return
    PRINT("### Removing access to auth0")
    remove_from_auth0_client(env)
    PRINT("### Deleting beanstalk enviornment")
    delete_bs_env_cli(env)
    PRINT("### Delete contents of s3")
    delete_s3_buckets(env)
    PRINT("### Delete es domain")
    delete_es_domain(env)
    PRINT("### Delete database")
    delete_db(env)
    PRINT('### All done!')
